# -*- coding: utf-8 -*-
__author__ = 'isparks'

accounts = {'innovate' : {'username': 'isparks',
                         'password':'Jupiter103'},
            'otherurl' : {'username': 'username',
                         'password':'password'},
           }